/*
    This file is a part of Musty, a simple Molecular Dynamics simulation
    package.
    
    Copyright (C) 2005 Jori Liesenborgs

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "mdsimulation.h"
#include "ljforce.h"
#include "uniformdistribution.h"
#include "constants.h"
#include <stdio.h>
#include <iostream>
#include <string>

#define HISTPOINTS 1000 // Number of points in the histogram used to calculate the 
                        // pair correlation function

class MyMDSim : public MDSimulation
{
public:
	MyMDSim(int numpart,int max,const std::string &filename,const std::string &comment) : MDSimulation(numpart) 	
	{ 
		maxsteps = max; 
		
		// A series of positions will be stored to calculate the diffusion coefficient
		// from the value of < (delta r)^2 >

		positions = new Vector3D*[max];

		// A series of velocities will be stored to calculate the velocity autocorrelation
		// function
		
		velocities = new Vector3D*[max]; 

		for (int i = 0 ; i < max ; i++) 
		{
			velocities[i] = new Vector3D[numpart]; 
			positions[i] = new Vector3D[numpart];
		}
		
		fname = filename;
		cmt = comment;

		// If writesim is set, the positions of the particles will be written
		// to a file called 'sim.gnuplot'. Using 'gnuplot sim.gnuplot', the
		// simulation can be visualized.
		writesim = false; 
	}
	
	~MyMDSim() 										
	{ 
		for (int i = 0 ; i < maxsteps ; i++) 
		{
			delete [] velocities[i]; 
			delete [] positions[i];
		}
		delete [] positions;
		delete [] velocities; 
	}

	void SetWriteSimulation(bool flag)							{ writesim = flag; }
protected:
	bool OnStart()
	{
		if (writesim)
			f = fopen("sim.gnuplot","w");

		// Initialize the simulation step counter and clear the
		// histogram information
		num = 0;
		for (int i = 0 ; i < HISTPOINTS ; i++) 
			radiushistogram[i] = 0; 
		totalhistcount = 0;
		return true;
	}

	bool OnStep(bool *stop)
	{
		int i;
		int numpart = GetNumberOfParticles();

		// If requested, write current positions too the gnuplot file
		
		if (writesim)
		{
			fprintf(f,"splot [0:%g][0:%g][0:%g] '-'\n",(double)GetBoxSize(),(double)GetBoxSize(),(double)GetBoxSize());
			for (i = 0 ; i < numpart ; i++)
			{
				Vector3D pos = GetParticles()[i].GetPosition();
				fprintf(f,"%g %g %g\n",(double)pos.GetX(),(double)pos.GetY(),(double)pos.GetZ());
			}
			fprintf(f,"e\n");
			fprintf(f,"pause 0.004\n");
		}
			
		if (!InEquilibrium())
			return true;

		// store velocities and positions

		if (num == 0)	
		{
			for (i = 0 ; i < numpart ; i++)
				positions[num][i] = GetParticles()[i].GetPosition();
		}
		else
		{
			// Here, the new positions are calculated from the
			// previous positions and the particle step. The particle
			// step is used to avoid problems with the periodic boundary
			// conditions
			
			for (i = 0 ; i < numpart ; i++)
				positions[num][i] = positions[num-1][i] + GetParticleSteps()[i];
		}
		
		for (i = 0 ; i < numpart ; i++)
			velocities[num][i] = GetParticles()[i].GetVelocity();
		
		// update histogram

		if (num%10 == 0) // every ten steps, the histogram is updated
		{
			real_t boxsize = GetBoxSize();
			real_t maxdist = boxsize/2.0;
			int j;
			
			for (i = 0 ; i < numpart ; i++)
			{
				for (j = i+1 ; j < numpart ; j++)
				{
					real_t dist = GetInterParticleVector(i,j).GetLength();

					if (dist < maxdist)
					{
						int index = (int)(((dist/maxdist)*((real_t)HISTPOINTS))+0.5);
						if (index >= 0 && index < HISTPOINTS)
						{
							radiushistogram[index]++;
						}
					}
				}
			}
			totalhistcount++;
		}

		if ((num%100) == 0)
			printf("step = %d\n",num);
		num++;
		if (num >= maxsteps)
			*stop = true;
		return true;
	}

	void OnStop(bool error)
	{
		if (writesim)
		{
			fprintf(f,"pause -1\n");
			fclose(f);
		}
		
		if (error)
		{
			std::cerr << "Error: exiting" << std::endl;
			return;
		}
		
		printf("Gathered data, analyzing...\n");
		
		int numpart = GetNumberOfParticles();
		FILE *g = fopen(fname.c_str(),"w");
		if (g == 0)
		{
			std::cerr << "Error: can't open file " << fname << std::endl;
			return;
		}

		// write some info about the simulation
		fprintf(g,"# %s\n",cmt.c_str());
		
		int numcvs = maxsteps/2;
		real_t *c_vs = new real_t[numcvs];
		
		// calculate velocity autocorrelation function
		
		for (int t = 0 ; t < numcvs ; t++)
		{
			real_t c_v = 0;

			for (int i = 0 ; i < numpart ; i++)
			{
				real_t term = 0;
				real_t val;

				for (int t2 = 0 ; t2 < numcvs ; t2++)
				{
					val = velocities[t2][i]*velocities[t2+t][i];
					term += val;
				}
				term /= (real_t)(numcvs);

				c_v += term;
			}
			c_v /= (real_t)numpart;
			c_vs[t] = c_v;
			
			fprintf(g,"%d %g\n",t,(double)(c_v/c_vs[0]));
		}

		// calculate and write the pair correlation function
		
		fprintf(g,"\n\n");
		for (int i = 0 ; i < HISTPOINTS ; i++)
		{
			real_t r,gr;
			real_t boxsize = GetBoxSize();
			real_t dr = (boxsize/2.0)/HISTPOINTS;

			r = (((real_t)i)+0.5)*dr;
			gr = 2.0*boxsize*boxsize*boxsize/((real_t)(numpart*(numpart-1)))*(  (((real_t)radiushistogram[i])/((real_t)totalhistcount))/(4.0*CONST_PI*r*r*dr));
			
			fprintf(g,"%g %g\n",r,(double)gr);
		}

		// calculate and write <(delta r)^2>
	
		fprintf(g,"\n\n");

		real_t r2a,r2b;
		real_t ta,tb;

		r2a = 0;
		r2b = 0;
		ta = 0;
		tb = 0;
		
		for (int t = 0 ; t < numcvs ; t++)
		{
			real_t r2 = 0;

			for (int i = 0 ; i < numpart ; i++)
			{
				real_t term = 0;
				real_t val;

				for (int t2 = 0 ; t2 < numcvs ; t2++)
				{
					Vector3D diff = positions[t2+t][i]-positions[t2][i];
					val = diff.GetLength2();
					term += val;
				}
				term /= (real_t)(numcvs);

				r2 += term;
			}
			r2 /= (real_t)numpart;
			
			if (t == numcvs/2)
			{
				r2a = r2;
				ta = ((real_t)t)*GetTimeStep();
			}
			else if (t == numcvs-1)
			{
				r2b = r2;
				tb = ((real_t)t)*GetTimeStep();
			}
			fprintf(g,"%d %g\n",t,(double)(r2));
		}
		
		// calculate diffustion coefficient from the velocity
		// autocorrelation function

		real_t D = 0;
		real_t dt = GetTimeStep();
		
		for (int t = 0 ; t < numcvs-1 ; t++)
			D += dt*(c_vs[t]+c_vs[t+1])/2.0;

		D /= 3.0;
		
		fprintf(g,"# D = %g (in used units)\n",(double)D);

		D *= GetLengthUnit()*GetLengthUnit()/GetTimeUnit();

		fprintf(g,"# D = %g (in SI units)\n",(double)D);

		delete [] c_vs;
	
		// use <(delta r)^2> to calculate another estimate
		// of the diffusion coefficient
		
		real_t D2 = (1.0/6.0)*(r2b-r2a)/(tb-ta);

		fprintf(g,"# D2 = %g (in used units)\n",(double)D2);

		D2 *= GetLengthUnit()*GetLengthUnit()/GetTimeUnit();

		fprintf(g,"# D2 = %g (in SI units)\n",(double)D2);

		fclose(g);
	}
	
private:
	int num,maxsteps;
	Vector3D **velocities;
	Vector3D **positions;
	FILE *f;
	std::string fname,cmt;
	int radiushistogram[HISTPOINTS];
	int totalhistcount;

	bool writesim;
};

int main(void)
{
	// Info for Argon
	
	real_t T0 = 119.8;
	real_t energyunit = T0*1.381e-23;
	real_t lengthunit = 3.822e-10;
	real_t massunit = 39.948*1.660538e-27;

	// Read some parameters

	int numparticles,steps;
	real_t rho,temperature;
	std::string fname;

	std::cout << "Enter number of particles: ";
	std::cin >> numparticles;
	std::cout << std::endl;
	std::cout << "Enter density (in units of 1/sigma^3): ";
	std::cin >> rho;
	std::cout << std::endl;
	std::cout << "Enter temperature (in units of epsilon/kB): ";
	std::cin >> temperature;
	std::cout << std::endl;
	std::cout << "Enter output file name: ";
	std::cin >> fname;
	std::cout << std::endl;
	
	std::cout << "Enter number of steps after equilibrium: ";
	std::cin >> steps;
	std::cout << std::endl;
	
	char info[1024];

	sprintf(info,"N = %d, rho = %g, T = %g, steps = %d",numparticles,(double)rho,(double)temperature,steps);
	
	MyMDSim sim(numparticles,steps,fname,std::string(info));
	LJForce ljforce(1.0,1.0);

	// sim.SetWriteSimulation(true);
	
	temperature *= T0;
	real_t boxlength = POW(((real_t)numparticles)/rho,1.0/3.0);

	// Run simulation

	if (sim.Run(ljforce,temperature,boxlength,0.004,lengthunit,massunit,energyunit,500))
	{
		std::cout << std::endl << std::endl;
		std::cout << "Output written to '" << fname << "'" << std::endl;
		std::cout << "In gnuplot, use:" << std::endl;
		std::cout << "  plot '" << fname << "' index 0 # to plot the velocity autocorrelation function" << std::endl;
		std::cout << "  plot '" << fname << "' index 1 # to plot the pair correlation function" << std::endl;
		std::cout << "  plot '" << fname << "' index 2 # to plot <(delta r)^2>" << std::endl;
		std::cout << "The last few lines contain D and D2, the diffusion coefficients determined from" << std::endl;
		std::cout << "the velocity autocorrelation function and <(delta r)^2> respectively. Type:" << std::endl;
		std::cout << "  tail '" << fname << "'" << std::endl;
		std::cout << "to display these results." << std::endl << std::endl;
	}
	
	return 0;
}

